## Distributed as part of the supplementary material for the the paper
## ``Improving the accuracy of likelihood-based inference in
## meta-analysis and meta-regression''
##
## Authors: Ioannis Kosmidis, Annamaria Guolo, Cristiano Varin
## Date: 15 October 2016
## Licence: GPL 2 or greater

## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!


## This script can been used to carry out the simulation study in
## Section 3.1 of the supplementary material and produce Figure 2 and
## Figure 3 therein

library("metaLik")
library("metatest")
library("parallel")

source("functionsMPL.R")

## as in Zhen and Lin...
generate.sigma2s <- function(K){
    sigma2s <- 0.25 * rchisq(K, df = 1)
    ok <- (sigma2s > 0.009 & sigma2s < 0.6)
    while(sum(ok) < K){
        tmp <- 0.25 * rchisq(K-sum(ok), df = 1)
        sigma2s[!ok] <- tmp
        ok <- (sigma2s > 0.009 & sigma2s < 0.6)
    }
    sigma2s
}

simulate.BG <- function(beta = 0.5,
                        var.re = 0.01,
                        sigma2s = runif(10, 0, 1)) {
    N <- length(sigma2s)
    list(betahat = beta + rnorm(N)*sqrt(var.re + sigma2s),
         betahat.se = sqrt(sigma2s))
}

## Number of cores to use for the simulation
ncores <- 7

## Direcory to store intermediate resuts
tmpdir <- "~/Downloads/"
prefix <- "simulations_BrokwellGordon_revision2_"

## Sample sizes to consider
Ks <- c(seq(5, 50, by = 5), 100, 200)

## Variance components to consider
truetaus <- sqrt(seq(0, 0.1, length = 11))
## Other simulation constants
truebeta <- 0.5

## Compute X matrices and generate sigma2
sigma2 <- X <- as.list(rep(0, length(Ks)))
## Generate sigmas and set Xs
set.seed(123)
for (j in seq_along(Ks)) {
    sigma2[[j]] <- generate.sigma2s(Ks[j])
    X[[j]] <- matrix(1, nrow = Ks[j])
}

## Simulation size per sample size
nsimu <- 10000
## Bootstrap size for double-resampling
B <- 1000


## Objects that should be in all images
basic_objects <- c("Ks", "truebeta", "truetaus", "nsimu", "B", "current_results")

## Simulation and computation of estimates and p-values
for (i in seq_along(Ks)) {
    ## Object to store simulation results
    current_results <- as.list(numeric(length(truetaus)))
    ## Save an image with the basic objects for K = Ks[i]
    image_name <- paste0(tmpdir, "/", prefix, i, "_K", Ks[i], ".rda")
    save(list = basic_objects, file = image_name)
    for (j in seq_along(truetaus)) {

        ## Simulation constants
        num.studies <- Ks[i]
        truetau <- truetaus[j]

        cat("---------------------------------\n")
        cat("Sample size", num.studies, "'true' variance", truetau^2,"\n")

        ## Compute bias and mean squared error for estimating delta and the rate
        ## of Type I error for testing H_0: delta = truedelta
        set.seed(123)
        ## Simulate under truedelta
        cat("Simulating data...")
        simuDat <- lapply(seq.int(nsimu), function(k) {
            dat <- simulate.BG(beta = truebeta,
                               var.re = truetau^2,
                               sigma2s = sigma2[[i]])
            data.frame(dat)
            ## attr(out, "id") <- k
            ## out
        })
        cat("Done\n")

        ## Use BiasFit to get estimates for each of the simulated data sets
        cat("Getting estimates and tests using the random effects meta-analysis model...")
        cres <- mclapply(seq.int(nsimu), function(id) {
            cdat <- simuDat[[id]]
            if (id %% 1000 == 0) {
                cat("Sample", id, "\n")
            }
            m1 <- metaLik(betahat ~ 1, data = cdat, sigma2 = betahat.se^2)
            list(estimates = simplify2array(BiasFit(m1)),
                 pvalues = perform_tests(m1$y, m1$X, m1$sigma2,
                                         null = truebeta, what = 1,
                                         B = B,
                                         silent = FALSE,
                                         maxiter = 10000)[, c("pvalues_l", "pvalues_d", "pvalues_g")])
        }, mc.cores = ncores)
        cat("Done\n")

        ## Load the image with the basic objects
        cat("Loading the image with the basic objects...")
        load(image_name)
        cat("Done\n")
        ## Put cres in the list
        cat("Allocating the estimates and tests...")
        current_results[[j]] <- cres
        cat("Done\n")
        ## Overwrite the image with the basic objects
        cat("Overwriting the basic objects in the image...")
        save(list = basic_objects, file = image_name)
        cat("Done\n")
        ## Delete current_results
        cat("Cleaning up...\n")
        rm(current_results)
        rm(simuDat)
        rm(cres)
        cat("Done\n")

        ## Collect garbage
        gc()

    }
}




## Compute moments and sizes and then combine images
## Clean-up
rm(list = ls())

compute_moments <- function(coefs, truth) {
    bias <- rowMeans(coefs - truth)
    mse <- rowMeans((coefs - truth)^2)
    data.frame(bias, mse)
}

alphas <- c(0.01, 0.025, 0.05, 0.1)

## Directory where intermediate resuts are
tmpdir <- "~/Desktop/combine/Individual results/"#Downloads/"
prefix <- "simulations_BrokwellGordon_revision2_"

## Ks covered by the results
K <- c(seq(5, 50, by = 5), 100, 200)

## Load first image to get basic objects
image_name <- paste0(tmpdir, "/", prefix, 1, "_K", K[1], ".rda")
load(image_name)
## The results list to populate
results <- as.list(numeric(length(K)))
names(results) <- K
for (i in seq_along(K)) {

    cat("Computing moments and sizes for K = ", K[i], "...")
    image_name <- paste0(tmpdir, "/", prefix, i, "_K", K[i], ".rda")
    load(image_name)

    ## Extract list of ML and MPL coefficients and calculate moments
    coefs <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$estimates[1, ]
        }))
    })
    moments <- lapply(coefs, compute_moments, truth = truebeta)


    ## Size for H_0: beta = null versus beta != null
    pvalues_d <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_d"]
        }))
    })
    size_d <- lapply(pvalues_d, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    ## Size for H_0: beta <= null versus beta > null
    pvalues_l <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_l"]
        }))
    })
    size_l <- lapply(pvalues_l, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    ## Size for H_0: beta >= null versus beta < null
    pvalues_g <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_g"]
        }))
    })
    size_g <- lapply(pvalues_g, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    results[[as.character(K[i])]] <- list(estimates_and_pvalues = current_results,
                                           size_d = size_d,
                                           size_g = size_g,
                                           size_l = size_l,
                                           moments = moments)
    cat("Done\n")
}
Ks <- K
save(list = c("results", "Ks", "truetaus"), file = paste0(tmpdir, "/", prefix, "all.rda"))




## Plots
## Clean-up
rm(list = ls())

## Directory where resuts are
tmpdir <- "~/Desktop/combine/"#Downloads/"
prefix <- "simulations_BrokwellGordon_revision2_"

load(paste0(tmpdir, "/", prefix, "all.rda"))
lev <- 0.95

## Plots as in supplementary
tau2s <- truetaus^2
names(tau2s) <- tau2s
names(Ks) <- Ks
n.methods <- 8 ## number methods under comparison!

alternative <- "different"

type <- switch(alternative,
               "different" = "size_d",
               "greater" = "size_g",
               "less" = "size_l")

## Plots versus psi for fixed K
pdf(paste0(tmpdir, "/", "BG_versus_psi_", type, ".pdf"), width = 10, height = 10*sqrt(2))

all.type1 <- array(dim = c(length(Ks), length(tau2s), n.methods))

for (j in seq_along(Ks)) {
    this.type1 <- t(simplify2array(lapply(results[[j]][[type]], function(x) x[, as.character(1 - lev)])))
    all.type1[j, ,] <- as.matrix(this.type1)
}

dimnames(all.type1) <- list(names(Ks), names(tau2s),  colnames(this.type1))

par(mfrow = c(6, 2))
par(mar = c(0.3, 0.3, 0.3, 0.3), oma = c(6, 6, 4, 4))
par(las = 1, cex.lab = 1.3, cex.axis = 1.3, cex.main = 1.3)
for (K in Ks[1:6]) {
    covers.a <- 1 - all.type1[as.character(K), , ] ## empirical coverage K = 10
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(psi), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"DL"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"ZL"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Skovgaard"] ~ tau2s, lwd = 2, lty = 4)
    tickPos <- c(0.85, 0.90, 0.95, 1)
    if (K == Ks[6]) {
        axis(1)
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    else {
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(psi), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"WaldML"] ~ tau2s, lwd = 2, lty = 2)
        lines(covers.a[ ,"LR"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Bartlett"] ~ tau2s, lwd = 2, lty = 4)
    if (K == Ks[6]) {
        axis(1)
    }
    mtext(bquote(K == .(K)), side = 4, adj = -0.1, cex = 0.9)
}
mtext("Coverage (%)", outer = TRUE, side = 2, cex = 1, las = 0, padj = -4)
mtext(expression(psi), outer = TRUE, side = 1, cex = 1, las = 0, padj = 4)
mtext(paste("Alternative:", alternative), outer = TRUE, side = 3, cex = 1, las = 0, padj = -2)
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), mar = c(0, 0, 0, 0), new = TRUE)
legend(x = 0.008, y = 1.006, inset = c(0, 0), legend = c("PLR", "DL", "DR", "Skov"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex= 1.1, bty = "n")
legend(x = 0.0544, y = 1.006, inset = c(0, 0), legend = c("PLR", "Wald", "LR", "bLR"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex = 1.1, bty = "n")

par(mfrow = c(6, 2))
par(mar = c(0.3, 0.3, 0.3, 0.3), oma = c(6, 6, 4, 4))
par(las = 1, cex.lab = 1.3, cex.axis = 1.3, cex.main = 1.3)
for (K in Ks[7:12]) {
    covers.a <- 1 - all.type1[as.character(K), , ] ## empirical coverage K = 10
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(psi), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"DL"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"ZL"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Skovgaard"] ~ tau2s, lwd = 2, lty = 4)
    tickPos <- c(0.85, 0.90, 0.95, 1)
    if (K == Ks[12]) {
        axis(1)
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    else {
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(psi), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"WaldML"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"LR"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Bartlett"] ~ tau2s, lwd = 2, lty = 4)
    if (K == Ks[12]) {
        axis(1)
    }
    mtext(bquote(K == .(K)), side = 4, adj = -0.1, cex = 0.9)
}
mtext("Coverage (%)", outer = TRUE, side = 2, cex = 1, las = 0, padj = -4)
mtext(expression(psi), outer = TRUE, side = 1, cex = 1, las = 0, padj = 4)
mtext(paste("Alternative:", alternative), outer = TRUE, side = 3, cex = 1, las = 0, padj = -2)
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), mar = c(0, 0, 0, 0), new = TRUE)
legend(x = 0.008, y = 1.006, inset = c(0, 0), legend = c("PLR", "DL", "DR", "Skov"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex= 1.1, bty = "n")
legend(x = 0.0544, y = 1.006, inset = c(0, 0), legend = c("PLR", "Wald", "LR", "bLR"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex = 1.1, bty = "n")
dev.off()

