## Distributed as part of the supplementary material for the the paper
## ``Improving the accuracy of likelihood-based inference in
## meta-analysis and meta-regression''
##
## Authors: Ioannis Kosmidis, Annamaria Guolo, Cristiano Varin
## Date: 15 October 2016
## Licence: GPL 2 or greater

## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

## This script can been used to carry out the simulation study in
## Section 3.3 of the supplementary material and produce Figure 6 and
## Figure 7 therein


library("metaLik")
library("metatest")
library("parallel")

source("functionsMPL.R")
source("simulator_IPDMA.R")

## Number of cores to use for the simulation
ncores <- 7

## Direcory to store intermediate resuts
tmpdir <- "~/metalik/"
prefix <- "simulations_IPDMA_revision2_mild_effect_"


## Sample sizes to consider
Ks <- c(seq(5, 50, by = 5), 100, 200)

## Variance components to consider
truesigmas <- sqrt(seq(0, 2.5, by = 0.25))

truealpha <- -1.27
truebeta <- 0.9
## truealpha <- -2
## truebeta <- 1.5
min.size <- 30
max.size <- 100
## covariate settings
binary.x <- TRUE
prevelance <- 0.5
mean.x <- 0
var.x <- 1
## Bias reduction for estimation?
br <- TRUE


## Simulation size per sample size
nsimu <- 10000
## Bootstrap size for double-resampling
B <- 1000

## Objects that should be in all images
basic_objects <- c("Ks", "truealpha", "truebeta", "min.size", "max.size", "binary.x", "prevelance", "mean.x", "var.x", "br", "truesigmas", "nsimu", "B",
                   "current_results")

for (i in seq_along(Ks)) {
    ## Object to store simulation results
    current_results <- as.list(numeric(length(truesigmas)))
    ## Save an image with the basic objects for K = Ks[i]
    image_name <- paste0(tmpdir, "/", prefix, i, "_K", Ks[i], ".rda")
    save(list = basic_objects, file = image_name)
    for (j in seq_along(truesigmas)) {

        ## Simulation constants
        num.studies <- Ks[i]
        truesigma <- truesigmas[j]

        cat("---------------------------------\n")
        cat("Sample size", Ks[i], "'true' variance", truesigmas[j]^2,"\n")

        ## Compute bias and mean squared error for estimating beta and the rate
        ## of Type I error for testing H_0: beta = truebeta
        set.seed(123)
        ## Simulate under truebeta
        cat("Simulating data...")
        simuDat <- mclapply(seq.int(nsimu), function(k) {
            set.seed(k)
            dat <- simulate.IPDMA(num.studies = num.studies,
                                  min.size = min.size,
                                  max.size = max.size,
                                  binary.x = binary.x,
                                  prevelance = prevelance,
                                  var.a = 0.1^2,
                                  var.b = truesigma^2,
                                  alpha = truealpha,
                                  beta = truebeta,
                                  br = br)
            data.frame(dat)
            ## attr(out, "id") <- k
            ## out
        }, mc.cores = ncores)
        cat("Done\n")

        ## Use BiasFit to get estimates for each of the simulated data sets
        cat("Getting estimates and tests using the random effects meta-analysis model...")
        cres <- mclapply(seq.int(nsimu), function(id) {
            cdat <- simuDat[[id]]
            if (id %% 1000 == 0) {
                cat("Sample", id, "\n")
            }
            m1 <- metaLik(betahat ~ 1, data = cdat, sigma2 = betahat.se^2)
            list(estimates = simplify2array(BiasFit(m1)),
                 pvalues = perform_tests(m1$y, m1$X, m1$sigma2,
                                         null = truebeta, what = 1,
                                         B = B,
                                         silent = FALSE,
                                         maxiter = 100000)[, c("pvalues_l", "pvalues_d", "pvalues_g")])
        }, mc.cores = ncores)
        cat("Done\n")

        ## Load the image with the basic objects
        cat("Loading the image with the basic objects...")
        load(image_name)
        cat("Done\n")
        ## Put cres in the list
        cat("Allocating the estimates and tests...")
        current_results[[j]] <- cres
        cat("Done\n")
        ## Overwrite the image with the basic objects
        cat("Overwriting the basic objects in the image...")
        save(list = basic_objects, file = image_name)
        cat("Done\n")
        ## Delete current_results
        cat("Cleaning up...\n")
        rm(current_results)
        rm(simuDat)
        rm(cres)
        cat("Done\n")

        ## Collect garbage
        gc()

    }
}



## Compute moments and sizes and then combine images
## Clean-up
rm(list = ls())

compute_moments <- function(coefs, truth) {
    bias <- rowMeans(coefs - truth)
    mse <- rowMeans((coefs - truth)^2)
    data.frame(bias, mse)
}

alphas <- c(0.01, 0.025, 0.05, 0.1)

## Directory where intermediate resuts are
tmpdir <- "~/Downloads/"
prefix <- "simulations_IPDMA_revision2_mild_effect_"

## Ks covered by the results
K <- c(seq(5, 50, by = 5), 100, 200)

## Load first image to get basic objects
image_name <- paste0(tmpdir, "/", prefix, 1, "_K", K[1], ".rda")
load(image_name)
## The results list to populate
results <- as.list(numeric(length(K)))
names(results) <- K
for (i in seq.int(K)) {

    cat("Computing moments and sizes for K = ", K[i], "...")
    image_name <- paste0(tmpdir, "/", prefix, i, "_K", K[i], ".rda")
    load(image_name)

    ## Extract list of ML and MPL coefficients and calculate moments
    coefs <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$estimates[1, ]
        }))
    })
    moments <- lapply(coefs, compute_moments, truth = truebeta)


    ## Size for H_0: beta = null versus beta != null
    pvalues_d <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_d"]
        }))
    })
    size_d <- lapply(pvalues_d, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    ## Size for H_0: beta <= null versus beta > null
    pvalues_l <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_l"]
        }))
    })
    size_l <- lapply(pvalues_l, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    ## Size for H_0: beta >= null versus beta < null
    pvalues_g <- lapply(current_results, function(x) {
        simplify2array(lapply(x, function(z) {
            z$pvalues[, "pvalues_g"]
        }))
    })
    size_g <- lapply(pvalues_g, function(x) {
        out <- sapply(alphas, function(a) {
            rowMeans(x < a, na.rm = TRUE)
        })
        colnames(out) <- alphas
        out
    })

    results[[as.character(K[i])]] <- list(estimates_and_pvalues = current_results,
                                          size_d = size_d,
                                          size_g = size_g,
                                          size_l = size_l,
                                          moments = moments)
    cat("Done\n")
}
Ks <- K
save(list = c("results", "Ks", "truealpha", "truebeta", "min.size", "max.size", "binary.x", "prevelance", "mean.x", "var.x", "br", "truesigmas"), file = paste0(tmpdir, "/", prefix, "all.rda"))


## Plots
## Clean-up
rm(list = ls())

## Directory where resuts are
tmpdir <- "~/Downloads/"
prefix <- "simulations_IPDMA_revision2_mild_effect_"

load(paste0(tmpdir, "/", prefix, "all.rda"))
lev <- 0.95

## Plots as in supplementary
tau2s <- truesigmas^2
names(tau2s) <- tau2s
names(Ks) <- Ks
n.methods <- 8 ## number methods under comparison!

alternative <- "different"

type <- switch(alternative,
               "different" = "size_d",
               "greater" = "size_g",
               "less" = "size_l")

## Plots versus psi for fixed K
pdf(paste0(tmpdir, "/", "IPDMA_mild_effect_versus_psi_", type, ".pdf"), width = 10, height = 10*sqrt(2))

all.type1 <- array(dim = c(length(Ks), length(tau2s), n.methods))

for (j in seq_along(Ks)) {
    this.type1 <- t(simplify2array(lapply(results[[j]][[type]], function(x) x[, as.character(1 - lev)])))
    all.type1[j, ,] <- as.matrix(this.type1)
}

dimnames(all.type1) <- list(names(Ks), names(tau2s),  colnames(this.type1))

par(mfrow = c(6, 2))
par(mar = c(0.3, 0.3, 0.3, 0.3), oma = c(6, 6, 4, 4))
par(las = 1, cex.lab = 1.3, cex.axis = 1.3, cex.main = 1.3)
for (K in Ks[1:6]) {
    covers.a <- 1 - all.type1[as.character(K), , ] ## empirical coverage K = 10
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(nu[2]), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"DL"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"ZL"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Skovgaard"] ~ tau2s, lwd = 2, lty = 4)
    tickPos <- c(0.85, 0.90, 0.95, 1)
    if (K == Ks[6]) {
        axis(1)
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    else {
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(nu[2]), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"WaldML"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"LR"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Bartlett"] ~ tau2s, lwd = 2, lty = 4)
    if (K == Ks[6]) {
        axis(1)
    }
    mtext(bquote(K == .(K)), side = 4, adj = -0.1, cex = 0.9)
}
mtext("Coverage (%)", outer = TRUE, side = 2, cex = 1, las = 0, padj = -4)
mtext(expression(nu[2]), outer = TRUE, side = 1, cex = 1, las = 0, padj = 4)
mtext(paste("Alternative:", alternative), outer = TRUE, side = 3, cex = 1, las = 0, padj = -2)
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), mar = c(0, 0, 0, 0), new = TRUE)
legend(x = 0.2, y = 1.006, inset = c(0, 0), legend = c("PLR", "DL", "DR", "Skov"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex= 1.1, bty = "n")
legend(x = 1.36, y = 1.006, inset = c(0, 0), legend = c("PLR", "Wald", "LR", "bLR"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex = 1.1, bty = "n")

par(mfrow = c(6, 2))
par(mar = c(0.3, 0.3, 0.3, 0.3), oma = c(6, 6, 4, 4))
par(las = 1, cex.lab = 1.3, cex.axis = 1.3, cex.main = 1.3)
for (K in Ks[7:12]) {
    covers.a <- 1 - all.type1[as.character(K), , ] ## empirical coverage K = 10
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(nu[2]), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"DL"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"ZL"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Skovgaard"] ~ tau2s, lwd = 2, lty = 4)
    tickPos <- c(0.85, 0.90, 0.95, 1)
    if (K == Ks[12]) {
        axis(1)
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    else {
        axis(2, at = tickPos, labels = tickPos * 100)
    }
    plot(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, ylim = c(0.85, 1), ylab = "", xlab = expression(nu[2]), type = "n", xaxt = "n", yaxt = "n")
    grid()
    abline(h = 0.95, lwd = 2, col = "darkgray")
    lines(covers.a[ ,"PLR"] ~ tau2s, lwd = 2, lty = 1)
    lines(covers.a[ ,"WaldML"] ~ tau2s, lwd = 2, lty = 2)
    lines(covers.a[ ,"LR"] ~ tau2s, lwd = 2, lty = 3)
    lines(covers.a[ ,"Bartlett"] ~ tau2s, lwd = 2, lty = 4)
    if (K == Ks[12]) {
        axis(1)
    }
    mtext(bquote(K == .(K)), side = 4, adj = -0.1, cex = 0.9)
}
mtext("Coverage (%)", outer = TRUE, side = 2, cex = 1, las = 0, padj = -4)
mtext(expression(nu[2]), outer = TRUE, side = 1, cex = 1, las = 0, padj = 4)
mtext(paste("Alternative:", alternative), outer = TRUE, side = 3, cex = 1, las = 0, padj = -2)
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), mar = c(0, 0, 0, 0), new = TRUE)
legend(x = 0.2, y = 1.006, inset = c(0, 0), legend = c("PLR", "DL", "DR", "Skov"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex= 1.1, bty = "n")
legend(x = 1.36, y = 1.006, inset = c(0, 0), legend = c("PLR", "Wald", "LR", "bLR"), lty = 1:4, xpd = TRUE, horiz = TRUE, cex = 1.1, bty = "n")
dev.off()

